"""
    Copyright (c) 2014-2023 Ad Schellevis <ad@opnsense.org>
    All rights reserved.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.

    THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
    INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
    AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
    AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
    OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
    SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
    INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
    CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
    ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
    POSSIBILITY OF SUCH DAMAGE.
"""
import syslog


def singleton(cls, *args, **kwargs):
    """ singleton pattern, use ad decorator
    :param cls:
    """
    instances = {}

    # noinspection PyShadowingNames
    def getinstance(*args, **kwargs):
        if cls not in instances:
            instances[cls] = cls(*args, **kwargs)
        return instances[cls]

    return getinstance


def emit_syslog(priority, message):
    msg = message.replace('\n', ' ')[:4000]
    syslog.syslog(priority, msg)


def syslog_debug(message):
    emit_syslog(syslog.LOG_DEBUG, message)


def syslog_notice(message):
    emit_syslog(syslog.LOG_NOTICE, message)


def syslog_info(message):
    emit_syslog(syslog.LOG_INFO, message)


def syslog_error(message):
    emit_syslog(syslog.LOG_ERR, message)


def syslog_auth_info(message):
    emit_syslog(syslog.LOG_AUTH | syslog.LOG_INFO, message)

def syslog_auth_error(message):
    emit_syslog(syslog.LOG_AUTH | syslog.LOG_ERR, message)
